
module cwx.editor.gui.dwt.images;

import cwx.utils;
import cwx.props;

import cwx.editor.gui.dwt.utils;
import cwx.editor.gui.dwt.props;

import std.math;
import std.file;
import std.path;

import dwt.DWT;
import dwt.DWTException;
import dwt.widgets.Display;
import dwt.widgets.Shell;
import dwt.widgets.Canvas;
import dwt.widgets.Composite;
import dwt.widgets.Control;
import dwt.widgets.Listener;
import dwt.widgets.Event;
import dwt.graphics.Color;
import dwt.graphics.Device;
import dwt.graphics.GC;
import dwt.graphics.Image;
import dwt.graphics.ImageData;
import dwt.graphics.Cursor;
import dwt.graphics.Font;
import dwt.graphics.FontData;
import dwt.graphics.Point;
import dwt.graphics.Rectangle;
import dwt.graphics.GCData;
import dwt.dnd.DND;
import dwt.dnd.DropTarget;
import dwt.dnd.DropTargetAdapter;
import dwt.dnd.DropTargetEvent;
import dwt.events.PaintListener;
import dwt.events.MouseAdapter;
import dwt.events.MouseListener;
import dwt.events.MouseMoveListener;
import dwt.events.DisposeListener;
import dwt.events.DisposeEvent;

/// イメージを拡大・縮小・移動させるためのトグル。
public enum Toggle {
	/// 左上のトグル。
	LEFT_TOP,
	/// 左辺中央のトグル。
	LEFT_MIDDLE,
	/// 左下のトグル。
	LEFT_BOTTOM,
	/// 上辺中央のトグル。
	MIDDLE_TOP,
	/// 下辺中央のトグル。
	MIDDLE_BOTTOM,
	/// 右上のトグル。
	RIGHT_TOP,
	/// 右辺中央のトグル。
	RIGHT_MIDDLE,
	/// 右下のトグル。
	RIGHT_BOTTOM,
	/// 画像内。(移動)
	MOVE,
	/// 画像外。
	NONE,
}

/// 画像を重ねて1枚のイメージを作成する。
public class PileImage {
private:
	public enum TPos {
		LEFT,
		RIGHT
	}
	struct AppImg {
		CInsets insets;
		string path = "";
		string text = "";
		ImageData data = null;
		bool transparent;
		int maskX;
		int maskY;
		CFont font;
		CRGB fontColor;
		TPos textPos = TPos.LEFT;
	}
	string _title = null;
	FontData titFont = null;
	Point titPoint = null;
	AppImg[] appends = [];
	Rectangle rect;
	bool t = false;
	bool s = false;
	Image _img;
	string path = "";
	ImageData data = null;

	bool _visible = true;

	int initW, initH;

public:
	/// 画像のファイルパス、位置、サイズを指定してインスタンスを生成する。
	/// パスが存在しない場合、描画のタイミングで単に表示されない。
	/// Params:
	/// path = 画像のファイルパス。
	/// x = 横位置。
	/// y = 縦位置。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(string path, int x, int y, int baseW, int baseH) {
		this.path = path;
		rect = new Rectangle(x, y, baseW, baseH);
		initW = baseW;
		initH = baseH;
	}
	/// 画像のファイルパス、サイズを指定してインスタンスを生成する。
	/// パスが存在しない場合、描画のタイミングで単に表示されない。
	/// Params:
	/// path = 画像のファイルパス。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(string path, int baseW, int baseH) {
		this(path, 0, 0, baseW, baseH);
	}
	/// 画像のデータ、位置、サイズを指定してインスタンスを生成する。
	/// パスが存在しない場合、描画のタイミングで単に表示されない。
	/// Params:
	/// data = 画像のデータ。
	/// x = 横位置。
	/// y = 縦位置。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(ImageData data, int x, int y, int baseW, int baseH) {
		this.data = data;
		rect = new Rectangle(x, y, baseW, baseH);
		initW = baseW;
		initH = baseH;
	}
	/// 画像のデータ、サイズを指定してインスタンスを生成する。
	/// パスが存在しない場合、描画のタイミングで単に表示されない。
	/// Params:
	/// data = 画像のデータ。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(ImageData data, int baseW, int baseH) {
		this(data, 0, 0, baseW, baseH);
	}
	/// Returns: ベースとなる幅。
	int baseWidth() {
		return initW;
	}
	/// Returns: ベースとなる高さ。
	int baseHeight() {
		return initH;
	}
	/// Params:
	/// initW = ベースとなる幅。
	void baseWidth(int initW) {
		this.initW = initW;
	}
	/// Params:
	/// initH = ベースとなる高さ。
	void baseHeight(int initH) {
		this.initH = initH;
	}

	/// 前面に画像を追加する。
	/// Params:
	/// path = 画像のファイルパス。
	/// insets = 内側の隙間。
	/// transparent = 透過するか否か。
	/// maskX = マスク色のX位置。
	/// maskY = マスク色のY位置。
	/// See_Also: createImage();
	void append(string path, CInsets insets, bool transparent, int maskX = 0, int maskY = 0) {
		AppImg append;
		append.insets = insets;
		append.path = path;
		append.transparent = transparent;
		append.maskX = maskX;
		append.maskY = maskY;
		appends ~= append;
	}
	/// ditto
	void append(ImageData data, CInsets insets) {
		AppImg append;
		append.insets = insets;
		append.data = data;
		appends ~= append;
	}
	/// ditto
	void append(ImageData data, CPoint point) {
		append(data, CInsets(point.y,
			initW - (point.x + data.width),
			initH - (point.y + data.height),
			point.x));
	}
	/// 前面に文字列を追加する。
	void append(string text, CInsets insets, CFont font, CRGB fontColor, TPos pos = TPos.LEFT) {
		AppImg append;
		append.text = text;
		append.insets = insets;
		append.font = font;
		append.fontColor = fontColor;
		append.textPos = pos;
		appends ~= append;
	}
	/// ditto
	void append(string text, CPoint point, CFont font, CRGB fontColor) {
		append(text, CInsets(point.y, 0, 0, point.x), font, fontColor, TPos.LEFT);
	}
	void setPath(string path) {
		this.path = path;
		this.data = null;
	}
	void setPath(int index, string path) {
		appends[index].path = path;
		appends[index].data = null;
	}
	void setTransparent(int index, bool mask) {
		appends[index].transparent = mask;
	}
	void setImageData(ImageData data) {
		this.path = "";
		this.data = data;
	}
	void setImageData(int index, ImageData data) {
		appends[index].path = "";
		appends[index].data = data;
	}
	/// タイトルとして表示する文字列を設定する。
	/// Params:
	/// title = タイトルの文字列。
	/// font = 表示時のフォント。
	/// titPoint = タイトルの表示位置。
	/// See_Also: createImage();
	void setTitle(string title, FontData font, Point titPoint) {
		this._title = title;
		this.titFont = font;
		this.titPoint = titPoint;
	}
	void title(string title) {
		assert (titFont);
		assert (titPoint);
		_title = title;
	}
	string title() {
		return this._title;
	}
	/// イメージ・タイトル・透明色の設定有無を設定した後に
	/// このメソッドを呼び出すことで、画像が生成される。
	/// See_Also: append(), setTitle(), transparent()
	void createImage() {
		if (_img) _img.dispose;
		auto data = createImageData;
		_img = data ? new Image(Display.getCurrent, data) : null;
	}
	/// イメージ・タイトル・透明色の設定有無を設定した後に
	/// このメソッドを呼び出すことで、ImageDataが生成される。
	/// See_Also: append(), setTitle(), transparent()
	ImageData createImageData() {
		if (width == 0 || height == 0 || initW == 0 || initH == 0) return null;
		auto cur = Display.getCurrent();
		auto bmp = new Image(cur, initW, initH);
		auto dc = new GC(bmp);

		ImageData matImgData;
		if (this.data) {
			matImgData = this.data;
		} else {
			if (isBinImg(path) || (path !is null && .exists(path))) {
				matImgData = loadImage(path, false).scaledTo(initW, initH);
			} else {
				// ファイルが無い場合は単に表示しない。
				matImgData = blankImage;
			}
		}
		auto matImg = new Image(cur, matImgData);
		dc.drawImage(matImg, 0, 0);
		matImg.dispose;

		foreach (a; appends) {
			if (a.path.length || a.data) {
				try {
					ImageData imgData;
					if (a.data) {
						imgData = a.data;
					} else {
						imgData = loadImage(a.path, a.transparent, a.maskX, a.maskY);
					}
					imgData = imgData.scaledTo
						(initW - a.insets.w - a.insets.e,
						initH - a.insets.n - a.insets.s);
					auto img = new Image(cur, imgData);
					scope (exit) img.dispose;
					dc.drawImage(img, a.insets.w, a.insets.n);
				} catch (DWTException e) {
					// ファイルが無い場合は表示しない。
				}
			}
			if (a.text.length) {
				try {
					auto font = new Font(cur, dwtData(a.font));
					scope (exit) font.dispose;
					dc.setFont = font;
					scope (exit) dc.setFont = null;
					int alpha;
					auto color = new Color(cur, dwtData(a.fontColor, alpha));
					scope (exit) color.dispose;
					auto fore = dc.getForeground;
					dc.setForeground = color;
					scope (exit) dc.setForeground = fore;
					dc.setAlpha = alpha;
					scope (exit) dc.setAlpha = 255;
					switch (a.textPos) {
					case TPos.LEFT: {
						dc.setTextAntialias = false;
						dc.drawText(a.text, a.insets.w, a.insets.n, true);
						dc.setTextAntialias = true;
					} break;
					case TPos.RIGHT: {
						int tw = dc.textExtent(a.text).x;
						dc.setTextAntialias = false;
						dc.drawText(a.text, initW - a.insets.e - tw, a.insets.n, true);
						dc.setTextAntialias = true;
					} break;
					default: assert (0);
					}
				} catch (DWTException e) {
				}
			}
		}

		if (_title !is null) {
			auto font = new Font(cur, titFont);
			dc.setFont = font;
			dc.setForeground(cur.getSystemColor(DWT.COLOR_BLACK));
			dc.setTextAntialias = false;
			dc.drawText(_title, titPoint.x, titPoint.y, true);
			dc.setTextAntialias = true;
			dc.setFont(null);
			font.dispose;
		}
		auto bmpData = bmp.getImageData;
		if (transparent) {
			bmpData.transparentPixel = bmpData.getPixel(0, 0);
		}
		bmp.dispose;
		dc.dispose;
		return bmpData.scaledTo(width, height);
	}
	/// 画像を描画する。
	/// Params:
	/// dc = キャンバス。
	void draw(GC gc) {
		if (_visible && _img) {
			gc.drawImage(_img, x, y);
		}
	}
	/// Returns: 表示するか。
	bool visible() {
		return _visible;
	}
	/// Params:
	/// v = 表示するか。
	void visible(bool v) {
		_visible = v;
	}
	/// Returns: 透明色を使用するか。
	bool transparent() {
		return t;
	}
	/// Params:
	/// t = 透明色を使用するか。
	/// See_Also: createImage();
	void transparent(bool t) {
		this.t = t;
	}
	/// Returns: 横位置。
	int x() {
		return rect.x;
	}
	/// 横位置を変更する。
	/// Params:
	/// x = 横位置。
	void x(int x) {
		rect.x = x;
	}
	/// Returns: 縦位置。
	int y() {
		return rect.y;
	}
	/// 縦位置を変更する。
	/// Params:
	/// y = 縦位置。
	void y(int y) {
		rect.y = y;
	}
	/// 幅を設定する。
	/// Params:
	/// w = 幅。
	void width(int w) {
		rect.width = w;
	}
	/// Returns: 幅。
	int width() {
		return rect.width;
	}

	/// 高さを設定する。
	/// Params:
	/// h = 高さ。
	void height(int h) {
		rect.height = h;
	}
	/// Returns: 高さ。
	int height() {
		return rect.height;
	}

	/// 位置とサイズを設定する。
	/// Params:
	/// rect = 位置とサイズ。
	void bounds(Rectangle rect) {
		this.rect.x = rect.x;
		this.rect.y = rect.y;
		this.rect.width = rect.width;
		this.rect.height = rect.height;
	}

	/// Returns: 位置とサイズ。
	Rectangle bounds() {
		return new Rectangle(x, y, width, height);
	}

	/// 全てのリソースを解放する。
	void dispose() {
		if (_img) _img.dispose;
	}
}

/// ImagePaneと組合わせて使用する移動可能な画像。
/// See_Also: ImagePane
public class FlexImage : PileImage {
private:
	void delegate(FlexImage img, int x, int y, real scale)[] lc_resizes;
	void delegate(FlexImage img, int x, int y, int w, int h)[] l_resizes;
	void delegate(FlexImage img)[] l_selected;

	int minW = 1, minH = 1;
	int maxW = 65536, maxH = 65536;
	bool s = false;
	bool whconst = false; // 縦横比を維持するか否か

	Rectangle newR;
	uint tglSize = 5;

	Rectangle[Toggle] tgls;
public:
	/// 画像のファイルパス、位置、本来のサイズを指定してインスタンスを生成する。
	/// パスが存在しない場合、描画のタイミングで単に表示されない。
	/// 指定されたパスが作成された場合、再描画の際に表示される。
	/// Params:
	/// path = 画像のファイルパス。
	/// x = 横位置。
	/// y = 縦位置。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(string path, int x, int y, int baseW, int baseH) {
		super(path, x, y, baseW, baseH);
		newR = new Rectangle(x, y, baseW, baseH);
	}
	/// 画像のデータ、位置、本来のサイズを指定してインスタンスを生成する。
	/// Params:
	/// path = 画像のデータ。
	/// x = 横位置。
	/// y = 縦位置。
	/// baseW = 本来の幅。
	/// baseH = 本来の高さ。
	this(ImageData data, int x, int y, int baseW, int baseH) {
		super(data, x, y, baseW, baseH);
		newR = new Rectangle(x, y, baseW, baseH);
	}
	/// Returns: 最小の幅。初期値は1。
	int minimumWidth() {
		return minW;
	}
	/// Params:
	/// minW = 最小の幅。
	void minimumWidth(int minW) {
		this.minW = minW;
	}
	/// Returns: 最小の高さ。初期値は1。
	int minimumHeight() {
		return minH;
	}
	/// Params:
	/// minW = 最小の高さ。
	void minimumHeight(int minH) {
		this.minH = minH;
	}
	/// Returns: 最大の幅。初期値は65536。
	int maximumWidth() {
		return maxW;
	}
	/// Params:
	/// minW = 最大の幅。
	void maximumWidth(int maxW) {
		this.maxW = maxW;
	}
	/// Returns: 最大の高さ。初期値は65536。
	int maximumHeight() {
		return maxH;
	}
	/// Params:
	/// minW = 最大の高さ。
	void maximumHeight(int maxH) {
		this.maxH = maxH;
	}
	/// Returns: 縦横比固定か。
	bool ratioFix() {
		return whconst;
	}
	/// Params:
	/// whconst = 縦横比固定ならtrue。
	void ratioFix(bool whconst) {
		this.whconst = whconst;
	}
	/// サイズ変更/移動更作業を終えてサイズ/位置を確定し、画像をその位置に配置する。
	/// 配置後、createImage()が実行される。
	/// See_Also: createImage();
	void resize() {
		bounds = newR;
		foreach (l; l_resizes) {
			l(this, x, y, width, height);
		}
		foreach (l; lc_resizes) {
			l(this, x, y, cast(real) width / initW);
		}
		createImage;
		retoggle;
	}
	/// 画像を描画する。
	/// Params:
	/// dc = キャンバス。
	void draw(GC gc) {
		if (visible) {
			super.draw(gc);

			if (selected) {
				gc.setBackground(Display.getCurrent.getSystemColor(DWT.COLOR_WHITE));
				gc.setForeground(Display.getCurrent.getSystemColor(DWT.COLOR_BLACK));
				gc.drawFocus(newX, newY, newWidth, newHeight);
				foreach (rect; tgls.values) {
					gc.fillRectangle(rect.x + 1, rect.y + 1, tglSize - 1, tglSize - 1);
					gc.drawRectangle(rect);
				}
			}
		}
	}
	/// Returns: 選択中か。
	bool selected() {
		return s;
	}
	/// Params:
	/// s = 選択状態。
	private void selected(bool s) {
		this.s = s;
	}
	private void doSelected(bool s) {
		selected = s;
		foreach (func; l_selected) {
			func(this);
		}
	}
	/// Returns: 仮の横位置。
	int newX() {
		return newR.x;
	}
	/// 横位置を仮に変更する。確定するにはresize()を使用。
	/// Params:
	/// x = 横位置。
	/// See_Also: resize()
	void newX(int x) {
		newR.x = x;
		retoggle();
	}
	/// Returns: 仮の縦位置。
	int newY() {
		return newR.y;
	}
	/// 縦位置を仮に変更する。確定するにはresize()を使用。
	/// Params:
	/// y = 縦位置。
	/// See_Also: resize()
	void newY(int y) {
		newR.y = y;
		retoggle();
	}
	/// 幅を設定可能な値に丸めて返す。
	/// Params:
	/// w = 幅。
	/// Returns: 丸めた幅。
	int roundWidth(int w) {
		w = minW > w ? minW : w;
		w = maxW < w ? maxW : w;
		return w;
	}
	/// Returns: 仮の幅。
	int newWidth() {
		return newR.width;
	}
	/// 幅を仮に設定する。確定するにはresize()を使用。
	/// Params:
	/// w = 幅。
	void newWidth(int w) {
		newR.width = roundWidth(w);
		retoggle();
	}
	/// 高さを設定可能な値に丸めて返す。
	/// Params:
	/// h = 高さ。
	/// Returns: 丸めた高さ。
	int roundHeight(int h) {
		if (whconst) {
			// 縦横比固定
			real scale = newWidth / cast(real) initW;
			return cast(int) rndtol(initH * scale);
		} else {
			h = minH > h ? minH : h;
			h = maxH < h ? maxH : h;
			return h;
		}
	}
	/// Returns: 仮の高さ。
	int newHeight() {
		return newR.height;
	}
	/// 高さを仮に設定する。確定するにはresize()を使用。
	/// Params:
	/// h = 高さ。
	void newHeight(int h) {
		newR.height = roundHeight(h);
		retoggle();
	}
	/// 位置とサイズを仮に設定する。確定するにはresize()を使用。
	/// Params:
	/// rect = 位置とサイズ。
	void newBounds(Rectangle rect) {
		newR.x = rect.x;
		newR.y = rect.y;
		newR.width = roundWidth(rect.width);
		newR.height = roundHeight(rect.height);
		retoggle();
	}
	/// スケールを指定してサイズを設定する。
	/// Params:
	/// scale = 元のサイズに対するスケール
	void scale(real scale) {
		newR.width = roundWidth(cast(int) rndtol(initW * scale));
		newR.height = roundHeight(cast(int) rndtol(initH * rect.height));
		retoggle();
	}
	/// Returns: 仮の位置とサイズ。
	Rectangle newBounds() {
		return new Rectangle(newR.x, newR.y, newR.width, newR.height);
	}

	/// 指定されたポイントが画像内、あるいはトグルの内部であればその値を返す。
	/// 画像外であればToggle.NONEを返す。
	/// Params:
	/// x = 横位置。
	/// y = 縦位置。
	/// Returns: トグル。
	Toggle inToggle(int x, int y) {
		foreach (key; tgls.keys) {
			auto rect = tgls[key];
			if (rect.x <= x && x <= (rect.x + rect.width)
					&& rect.y <= y && y <= (rect.y + rect.height)) {
				return key;
			}
		}
		if (this.x <= x && x <= (this.x + this.width)
				&& this.y <= y && y <= (this.y + this.height)) {
			return Toggle.MOVE;
		} else {
			return Toggle.NONE;
		}
	}

	/// 自在にサイズ変更できるときのトグル。
	Toggle[] RESIZE_TOGGLES = [
		Toggle.LEFT_TOP, Toggle.LEFT_MIDDLE, Toggle.LEFT_BOTTOM,
		Toggle.MIDDLE_TOP, Toggle.MIDDLE_BOTTOM,
		Toggle.RIGHT_TOP, Toggle.RIGHT_MIDDLE, Toggle.RIGHT_BOTTOM
	];
	/// 縦横比固定時のトグル。
	Toggle[] RESIZE_TOGGLES_CORNER = [
		Toggle.LEFT_TOP, Toggle.LEFT_BOTTOM,
		Toggle.RIGHT_TOP, Toggle.RIGHT_BOTTOM
	];

	private void retoggle() {
		Toggle[] tgls = whconst ? RESIZE_TOGGLES_CORNER : RESIZE_TOGGLES;
		foreach (key; this.tgls.keys) {
			this.tgls.remove(key);
		}
		foreach (tgl; tgls) {
			this.tgls[tgl] = toggleRect(tgl);
		}
	}

	private Rectangle toggleRect(Toggle tgl) {
		int tglX;
		int tglY;
		int tglWidth = tglSize;
		int tglHeight = tglSize;
		switch (tgl) {
		case Toggle.LEFT_TOP, Toggle.LEFT_MIDDLE, Toggle.LEFT_BOTTOM:
			tglX = newX - tglSize + (tglSize / 2);
			break;
		case Toggle.MIDDLE_TOP, Toggle.MIDDLE_BOTTOM:
			tglX = newX + (newWidth / 2) - (tglSize / 2);
			break;
		case Toggle.RIGHT_TOP, Toggle.RIGHT_MIDDLE, Toggle.RIGHT_BOTTOM:
			tglX = newX + newWidth - (tglSize / 2) - 1;
			break;
		}
		switch (tgl) {
		case Toggle.LEFT_TOP, Toggle.MIDDLE_TOP, Toggle.RIGHT_TOP:
			tglY = newY - tglSize + (tglSize / 2);
			break;
		case Toggle.LEFT_MIDDLE, Toggle.RIGHT_MIDDLE:
			tglY = newY + (newHeight / 2) - (tglSize / 2);
			break;
		case Toggle.LEFT_BOTTOM, Toggle.RIGHT_BOTTOM, Toggle.MIDDLE_BOTTOM:
			tglY = newY + newHeight - (tglSize / 2) - 1;
			break;
		}
		return new Rectangle(tglX, tglY, tglWidth, tglHeight);
	}

	/// 移動後に描画する領域を返す。
	/// Returns: 描画する領域。
	Rectangle drawNewArea() {
		return new Rectangle(newR.x - tglSize, newR.y - tglSize,
			newR.width + tglSize * 2, newR.height + tglSize * 2);
	}
	/// 描画する領域を返す。
	/// Returns: 描画する領域。
	Rectangle drawArea() {
		auto r = bounds;
		return new Rectangle(r.x - tglSize, r.y - tglSize,
			r.width + tglSize * 2, r.height + tglSize * 2);
	}

	/// リサイズの確定時に呼び出す関数を追加する。
	/// Params:
	/// サイズ確定時に呼び出す関数。
	void addResizeListener(void delegate(FlexImage img, int x, int y, int w, int h) func) {
		l_resizes ~= func;
	}
	/// ditto
	void addResizeListener(void delegate(FlexImage img, int x, int y, real scale) func) {
		lc_resizes ~= func;
	}
	/// 選択状態変更時に呼び出す関数を追加する。
	/// Params:
	/// 選択状態変更時に呼び出す関数。
	void addSelectionListener(void delegate(FlexImage img) func) {
		l_selected ~= func;
	}

	/// 全てのリソースを解放する。
	void dispose() {
		super.dispose;
	}
}

/// FlexImageと組合わせて柔軟に操作可能な画像を表示するパネル。
/// See_Also: FlexImage
public class ImagePane : Canvas {
private:
	/// トグルごとのカーソル。
	Cursor[Toggle] toggleCursors;

	/// トグルに応じたカーソルを返す。
	/// Params:
	/// tgl = トグル。
	/// Returns:
	/// サイズ変更が行える場合はその方向を示すカーソル、
	/// 移動が行える場合は手型のカーソル、それ以外の場合は通常のカーソル。
	Cursor getToggleCursor(Toggle tgl) {
		return toggleCursors[tgl];
	}

	Rectangle[FlexImage] dragImgs;
	Toggle dragTgl = Toggle.NONE;
	int dragStartX, dragStartY;
	bool moved = false;

	PileImage[] backs = [];

	class DListener : DisposeListener {
		public override void widgetDisposed(DisposeEvent e)  {
			foreach (img; backs) {
				img.dispose;
			}
		}
	}
	class MMListener : MouseMoveListener {
		override void mouseMove(MouseEvent me) {
			int x = me.x;
			int y = me.y;
			if (dragTgl != Toggle.NONE) {
				if (_ctrl && _mouseP) {
					doSelect(_mouseP);
				}
				int movX = x - dragStartX;
				int movY = y - dragStartY;
				foreach (pimg; dragImgs.keys) {
					if (cast(FlexImage) pimg) {
						auto img = cast(FlexImage) pimg;
						if (img.selected) {
							auto rect = dragImgs[img];
							Rectangle newRect = new Rectangle(img.x, img.y, img.width, img.height);
							switch (dragTgl) {
							case Toggle.LEFT_TOP, Toggle.LEFT_MIDDLE, Toggle.LEFT_BOTTOM:
								int w = rect.width - movX;
								newRect.width = img.roundWidth(w);
								break;
							case Toggle.RIGHT_TOP, Toggle.RIGHT_MIDDLE, Toggle.RIGHT_BOTTOM:
								int w = rect.width + movX;
								newRect.width = img.roundWidth(w);
								break;
							default:
								break;
							}
							switch (dragTgl) {
							case Toggle.LEFT_TOP, Toggle.MIDDLE_TOP, Toggle.RIGHT_TOP:
								int h = rect.height - movY;
								newRect.height = img.roundHeight(h);
								break;
							case Toggle.LEFT_BOTTOM, Toggle.MIDDLE_BOTTOM, Toggle.RIGHT_BOTTOM:
								int h = rect.height + movY;
								newRect.height = img.roundHeight(h);
								break;
							default:
								break;
							}
							switch (dragTgl) {
							case Toggle.LEFT_TOP, Toggle.MIDDLE_TOP, Toggle.RIGHT_TOP:
								newRect.y = rect.y + (rect.height - newRect.height);
								break;
							case Toggle.MOVE:
								newRect.y = rect.y + movY;
								break;
							default:
								break;
							}
							switch (dragTgl) {
							case Toggle.LEFT_TOP, Toggle.LEFT_MIDDLE, Toggle.LEFT_BOTTOM:
								newRect.x = rect.x + (rect.width - newRect.width);
								break;
							case Toggle.MOVE:
								newRect.x = rect.x + movX;
								break;
							default:
								break;
							}
							scope oldArea = img.drawNewArea;
							img.newBounds = newRect;
							scope newArea = img.drawNewArea;
							redraw(oldArea.x, oldArea.y, oldArea.width, oldArea.height, false);
							redraw(newArea.x, newArea.y, newArea.width, newArea.height, false);
						}
					}
				}
				moved = true;
			} else {
				foreach_reverse (pimg; backs) {
					if (cast(FlexImage) pimg && pimg.visible) {
						auto img = cast(FlexImage) pimg;
						Toggle tgl = img.inToggle(x, y);
						if (tgl != Toggle.NONE) {
							setCursor = getToggleCursor(tgl);
							return;
						}
					}
				}
				setCursor = getToggleCursor(Toggle.NONE);
			}
		}
	}

	class MouseDown : Listener {
		override void handleEvent(Event me) {
			setFocus;
			moved = false;
			_ctrl = (me.stateMask & DWT.SHIFT) != 0 || (me.stateMask & DWT.CTRL) != 0;
			int x = me.x;
			int y = me.y;
			if (me.button == 1) {
				dragStartX = x;
				dragStartY = y;
				auto tgl = Toggle.NONE;
				FlexImage img = null;
				foreach_reverse (i, pimg; backs) {
					if (cast(FlexImage) pimg) {
						img = cast(FlexImage) pimg;
						if (img.visible) {
							tgl = img.inToggle(x, y);
							if (tgl !is Toggle.NONE) {
								break;
							}
						}
					}
				}
				if (tgl !is Toggle.NONE) {
					_mouseP = img;
					if (me.button == 1) {
						if (!_ctrl) {
							if (!img.selected) doDeselectAll;
							doSelect(img);
						}
						dragTgl = tgl;
					} else if (me.button == 3) {
						if (!img.selected) doDeselectAll;
						doSelect(img);
					}
				} else {
					doDeselectAll;
					_mouseP = null;
				}
			}
		}
	}
	class FocusLost : Listener {
		override void handleEvent(Event me) {
			dragTgl = Toggle.NONE;
		}
	}
	class MouseUp : Listener {
		override void handleEvent(Event me) {
			int x = me.x;
			int y = me.y;
			if (me.button == 1) {
				bool ci = false;
				foreach_reverse (i, pimg; backs) {
					auto img = cast(FlexImage) pimg;
					if (img && img.selected) {
						scope oldRect = img.bounds;
						if (oldRect != img.newBounds) {
							if (!ci) {
								changingImages;
								ci = true;
							}
							img.resize;
							scope area = img.drawNewArea;
							redraw(oldRect.x, oldRect.y, oldRect.width, oldRect.height, false);
							redraw(area.x, area.y, area.width, area.height, false);
							dragImgs[img] = new Rectangle(img.x, img.y, img.width, img.height);
						}
					}
				}
				if (!moved && _mouseP) {
					if (_ctrl) {
						if (_mouseP.selected) {
							doDeselect(_mouseP);
						} else {
							doSelect(_mouseP);
						}
					} else {
						doDeselectAll;
						doSelect(_mouseP);
					}
				}
			}
			dragTgl = Toggle.NONE;
			moved = false;
			_mouseP = null;
		}
	}
	bool _ctrl = false;
	FlexImage _mouseP = null;
	void __setSelected(FlexImage img) {
		if (img.selected) {
			dragImgs[img] = new Rectangle(img.x, img.y, img.width, img.height);
			auto area = img.drawArea;
			redraw(area.x, area.y, area.width, area.height, false);
		}
	}
	class PListener : PaintListener {
		override void paintControl(PaintEvent e) {
			auto buf = new Image(getShell.getDisplay, getSize.x, getSize.y);
			auto gc = new GC(buf);

			gc.setBackground(getShell.getDisplay.getSystemColor(DWT.COLOR_DARK_BLUE));
			gc.fillRectangle(0, 0, getSize.x, getSize.y);
			foreach (bmp; backs) {
				bmp.draw(gc);
			}

			e.gc.drawImage(buf, 0, 0);
			buf.dispose;
		}
	}
public:

	PileImage[] images() {
		return backs;
	}

	void swap(int index1, int index2) {
		auto temp = backs[index1];
		backs[index1] = backs[index2];
		backs[index2] = temp;
	}

	void select(FlexImage[] imgs) {
		foreach (img; imgs) {
			select(img);
		}
	}
	void select(FlexImage img) {
		if (!img.selected) {
			img.selected = true;
			__setSelected(img);
		}
	}
	void select(int[] indices) {
		deselectAll;
		foreach (i; indices) {
			auto fi = cast(FlexImage) backs[i];
			select(fi);
		}
	}
	private void doSelect(FlexImage img) {
		if (!img.selected) {
			img.doSelected = true;
			__setSelected(img);
		}
	}
	void deselect(FlexImage img) {
		if (img.selected) {
			img.selected = false;
			deselAfter(img);
		}
	}
	private void doDeselect(FlexImage img) {
		if (img.selected) {
			img.doSelected = false;
			deselAfter(img);
		}
	}
	private void deselAfter(FlexImage img) {
		removeDragImage(img);
		auto area = img.drawArea;
		redraw(area.x, area.y, area.width, area.height, false);
	}
	private void doDeselectAll() {
		foreach_reverse (pimg; backs) {
			auto img = cast(FlexImage) pimg;
			if (img) doDeselect(img);
		}
	}
	void deselectAll() {
		foreach_reverse (pimg; backs) {
			auto img = cast(FlexImage) pimg;
			if (img) deselect(img);
		}
	}
	void deselectRange(int from, int to) {
		for (int i = from; i < to; i++) {
			auto img = cast(FlexImage) backs[i];
			if (img) deselect(img);
		}
	}
	int[] selectedIndices() {
		int[] r;
		foreach (i, img; backs) {
			auto fi = cast(FlexImage) img;
			if (fi && fi.selected) r ~= i;
		}
		return r;
	}
	int selectedIndex() {
		foreach_reverse (i, img; backs) {
			auto fi = cast(FlexImage) img;
			if (fi && fi.selected) return i;
		}
		return -1;
	}
	int findIndex(int x, int y) {
		foreach_reverse (i, img; backs) {
			auto fi = cast(FlexImage) img;
			if (fi && fi.visible && fi.bounds.contains(x, y)) return i;
		}
		return -1;
	}
	int findSelectedIndex(int x, int y) {
		foreach_reverse (i, img; backs) {
			auto fi = cast(FlexImage) img;
			if (fi && fi.visible && fi.selected && fi.bounds.contains(x, y)) return i;
		}
		return -1;
	}
	int[] findSelectedIndices(int x, int y) {
		int[] r;
		foreach (i, img; backs) {
			auto fi = cast(FlexImage) img;
			if (fi && fi.visible && fi.selected && fi.bounds.contains(x, y)) r ~= i;
		}
		return r;
	}

	private void removeDragImage(PileImage img) {
		auto fi = cast(FlexImage) img;
		if (fi && (fi in dragImgs)) {
			dragImgs.remove(fi);
		}
	}

	void remove(int index) {
		backs[index].dispose;
		removeDragImage(backs[index]);
		backs = backs[0 .. index] ~ backs[index + 1 .. $];
	}

	void removeRange(int fromIndex, int toIndex) {
		for (int i = fromIndex; i < toIndex; i++) {
			backs[i].dispose;
			removeDragImage(backs[i]);
		}
		backs = backs[0 .. fromIndex] ~ backs[toIndex .. $];
	}

	void insert(int index, PileImage img) {
		if (index == backs.length) {
			append(img);
		} else {
			backs = backs[0 .. index] ~ img ~ backs[index .. $];
			if (cast(FlexImage) img) __setSelected(cast(FlexImage) img);
		}
	}
	void insert(int index, PileImage[] imgs) {
		if (index == backs.length) {
			append(imgs);
		} else {
			backs = backs[0 .. index] ~ imgs ~ backs[index .. $];
			foreach (img; imgs) {
				if (cast(FlexImage) img) __setSelected(cast(FlexImage) img);
			}
		}
	}
	void set(int index, PileImage img) {
		backs[index].dispose;
		removeDragImage(backs[index]);
		this.backs[index] = img;
		if (cast(FlexImage) img) __setSelected(cast(FlexImage) img);
	}
	void append(PileImage img) {
		this.backs ~= img;
		if (cast(FlexImage) img) __setSelected(cast(FlexImage) img);
	}
	void append(PileImage[] imgs) {
		this.backs ~= imgs;
		foreach (img; imgs) {
			if (cast(FlexImage) img) __setSelected(cast(FlexImage) img);
		}
	}

	private void delegate()[] _changingImages;
	void changingImages(void delegate() changingImages) {
		_changingImages ~= changingImages;
	}
	private void changingImages() {
		foreach (ci; _changingImages) ci();
	}

	/// 唯一のコンストラクタ。
	this(Composite parent, int style) {
		super(parent, style);
		addListener(DWT.MouseDown, new MouseDown);
		addListener(DWT.MouseUp, new MouseUp);
		addMouseMoveListener(new MMListener);
		addListener(DWT.FocusOut, new FocusLost);
		addPaintListener(new PListener);
		addDisposeListener(new DListener);

		toggleCursors[Toggle.LEFT_TOP] = display.getSystemCursor(DWT.CURSOR_SIZENWSE);
		toggleCursors[Toggle.RIGHT_BOTTOM] = display.getSystemCursor(DWT.CURSOR_SIZENWSE);
		toggleCursors[Toggle.RIGHT_TOP] = display.getSystemCursor(DWT.CURSOR_SIZENESW);
		toggleCursors[Toggle.LEFT_BOTTOM] = display.getSystemCursor(DWT.CURSOR_SIZENESW);
		toggleCursors[Toggle.LEFT_MIDDLE] = display.getSystemCursor(DWT.CURSOR_SIZEWE);
		toggleCursors[Toggle.RIGHT_MIDDLE] = display.getSystemCursor(DWT.CURSOR_SIZEWE);
		toggleCursors[Toggle.MIDDLE_TOP] = display.getSystemCursor(DWT.CURSOR_SIZENS);
		toggleCursors[Toggle.MIDDLE_BOTTOM] = display.getSystemCursor(DWT.CURSOR_SIZENS);
		toggleCursors[Toggle.MOVE] = null;
		toggleCursors[Toggle.NONE] = null;
	}
}
