
module cwx.types;

import cwx.utils;

public:

/// 効果関連の例外。
class MotionException : Exception {
public:
	this(string msg) {
		super(msg);
	}
}

/// 精神状態。
enum Mentality {
	NORMAL, /// 正常。
	SLEEP, /// 睡眠。
	CONFUSE, /// 混乱。
	OVERHEAT, /// 激昂。
	BRAVE, /// 勇敢。
	PANIC /// 恐慌。
}
/// ditto
Mentality toMentality(string s) {
	switch (s) {
	case "Normal": return Mentality.NORMAL;
	case "Panic": return Mentality.PANIC;
	case "Brave": return Mentality.BRAVE;
	case "Overheat": return Mentality.OVERHEAT;
	case "Confuse": return Mentality.CONFUSE;
	case "Sleep": return Mentality.SLEEP;
	default: throw new Exception("Unknown Mentality: " ~ s);
	}
}
/// ditto
string fromMentality(Mentality m) {
	switch (m) {
	case Mentality.NORMAL: return "Normal";
	case Mentality.PANIC: return "Panic";
	case Mentality.BRAVE: return "Brave";
	case Mentality.OVERHEAT: return "Overheat";
	case Mentality.CONFUSE: return "Confuse";
	case Mentality.SLEEP: return "Sleep";
	}
}

/// 効果属性。
enum EffectType {
	PHYSIC, /// 物理。
	MAGIC, /// 魔法。
	MAGICAL_PHYSIC, /// 魔法的物理。
	PHYSICAL_MAGIC, /// 物理的魔法。
	NONE, /// 無。
}
/// 文字列から効果属性を生成。
EffectType toEffectType(string name) {
	switch (name) {
	case "Physic":
		return EffectType.PHYSIC;
	case "Magic":
		return EffectType.MAGIC;
	case "MagicalPhysic":
		return EffectType.MAGICAL_PHYSIC;
	case "PhysicalMagic":
		return EffectType.PHYSICAL_MAGIC;
	case "None":
		return EffectType.NONE;
	default:
		throw new MotionException("Unknown effecttype: " ~ name);
	}
}
/// 効果属性を文字列に変換。
string fromEffectType(EffectType etyp) {
	switch (etyp) {
	case EffectType.PHYSIC:
		return "Physic";
	case EffectType.MAGIC:
		return "Magic";
	case EffectType.MAGICAL_PHYSIC:
		return "MagicalPhysic";
	case EffectType.PHYSICAL_MAGIC:
		return "PhysicalMagic";
	case EffectType.NONE:
		return "None";
	}
}
/// 抵抗属性。
enum Resist {
	AVOID, /// 回避。
	RESIST, /// 抵抗。
	UNFAIL, /// 必中。
}
/// 文字列から抵抗属性を生成。
Resist toResist(string name) {
	switch (name) {
	case "Avoid":
		return Resist.AVOID;
	case "Resist":
		return Resist.RESIST;
	case "Unfail":
		return Resist.UNFAIL;
	default:
		throw new MotionException("Unknown resist: " ~ name);
	}
}
/// 抵抗属性から文字列へ変換。
string fromResist(Resist resist) {
	switch (resist) {
	case Resist.AVOID:
		return "Avoid";
	case Resist.RESIST:
		return "Resist";
	case Resist.UNFAIL:
		return "Unfail";
	}
}
/// 視覚効果。
enum CardVisual {
	NONE, /// 無し。
	REVERSE, /// 反転。
	HORIZONTAL, /// 横震動。
	VERTICAL, /// 縦振動。
}
/// 文字列から視覚効果を生成。
CardVisual toCardVisual(string name) {
	switch (name) {
	case "None":
		return CardVisual.NONE;
	case "Reverse":
		return CardVisual.REVERSE;
	case "Horizontal":
		return CardVisual.HORIZONTAL;
	case "Vertical":
		return CardVisual.VERTICAL;
	default:
		throw new MotionException("Unknown cardvisual: " ~ name);
	}
}
/// 視覚効果から文字列へ変換。
string fromCardVisual(CardVisual vis) {
	switch (vis) {
	case CardVisual.NONE:
		return "None";
	case CardVisual.REVERSE:
		return "Reverse";
	case CardVisual.HORIZONTAL:
		return "Horizontal";
	case CardVisual.VERTICAL:
		return "Vertical";
	}
}
/// 効果属性。
enum Element {
	ALL, /// 全。
	HEALTH, /// 肉体。
	MIND, /// 精神。
	MIRACLE, /// 神聖。
	MAGIC, /// 魔法。
	FIRE, /// 炎。
	ICE, /// 冷気。
}
/// 文字列から効果属性を生成。
Element toElement(string name) {
	switch (name) {
	case "All":
		return Element.ALL;
	case "Health":
		return Element.HEALTH;
	case "Mind":
		return Element.MIND;
	case "Miracle":
		return Element.MIRACLE;
	case "Magic":
		return Element.MAGIC;
	case "Fire":
		return Element.FIRE;
	case "Ice":
		return Element.ICE;
	default:
		throw new MotionException("Unknown element: " ~ name);
	}
}
/// 効果属性から文字列へ変換。
string fromElement(Element el) {
	switch (el) {
	case Element.ALL:
		return "All";
	case Element.HEALTH:
		return "Health";
	case Element.MIND:
		return "Mind";
	case Element.MIRACLE:
		return "Miracle";
	case Element.MAGIC:
		return "Magic";
	case Element.FIRE:
		return "Fire";
	case Element.ICE:
		return "Ice";
	}
}
/// 効果計算。
enum DamageType {
	LEVEL_RATIO, /// レベル比。
	NORMAL, /// 固定値。
	MAX, /// 最大値。
}
/// 文字列から効果計算方式を生成。
DamageType toDamageType(string name) {
	switch (name) {
	case "LevelRatio":
		return DamageType.LEVEL_RATIO;
	case "Normal":
		return DamageType.NORMAL;
	case "Max":
		return DamageType.MAX;
	default:
		throw new MotionException("Unknown damagetype: " ~ name);
	}
}
/// 効果計算方式を文字列へ変換。
string fromDamageType(DamageType dtyp) {
	switch (dtyp) {
	case DamageType.LEVEL_RATIO:
		return "LevelRatio";
	case DamageType.NORMAL:
		return "Normal";
	case DamageType.MAX:
		return "Max";
	}
}

/// メンバの選択。
struct Target {
public:
	/// メンバ種別。
	enum M {
		SELECTED, /// 選択中メンバ。
		UNSELECTED, /// 非選択メンバ。
		RANDOM, /// ランダムメンバ。
		PARTY, /// 全員。
	}
	M m; /// メンバ種別。
	bool sleep; /// 睡眠時有効可否。
	/// Targetを生成する。
	static Target opCall(M m, bool sleep) {
		Target r;
		r.m = m;
		r.sleep = sleep;
		return r;
	}
	int opEquals(Target t) {
		return t.m == m && t.sleep == sleep;
	}
private:
}
/// 文字列から対象メンバを生成。
Target toTarget(string name) {
	switch (name) {
	case "Selected":
		return Target(Target.M.SELECTED, false);
	case "Unselected":
		return Target(Target.M.UNSELECTED, false);
	case "Random":
		return Target(Target.M.RANDOM, false);
	case "Party":
		return Target(Target.M.PARTY, false);
	case "SelectedSleep":
		return Target(Target.M.SELECTED, true);
	case "UnselectedSleep":
		return Target(Target.M.UNSELECTED, true);
	case "RandomSleep":
		return Target(Target.M.RANDOM, true);
	case "PartySleep":
		return Target(Target.M.PARTY, true);
	default:
		throw new MotionException("Unknown targetm: " ~ name);
	}
}
/// 対象メンバを文字列へ変換。
string fromTarget(Target targ) {
	string targetText(string text, bool sleep) {
		return sleep ? text ~ "Sleep" : text;
	}
	switch (targ.m) {
	case Target.M.SELECTED:
		return targetText("Selected", targ.sleep);
	case Target.M.UNSELECTED:
		return targetText("Unselected", targ.sleep);
	case Target.M.RANDOM:
		return targetText("Random", targ.sleep);
	case Target.M.PARTY:
		return targetText("Party", targ.sleep);
	}
}

/// 精神要素。
enum Mental {
	AGGRESSIVE,/// 好戦
	UNAGGRESSIVE,/// 平和
	CHEERFUL, /// 社交
	UNCHEERFUL, /// 内向
	BRAVE, /// 勇敢
	UNBRAVE, /// 臆病
	CAUTIOUS, /// 慎重
	UNCAUTIOUS, /// 大胆
	TRICKISH, /// 狡猾
	UNTRICKISH, /// 正直
}
/// 文字列から精神要素を生成。
Mental toMental(string name) {
	switch (name) {
	case "Aggressive":
		return Mental.AGGRESSIVE;
	case "Unaggressive":
		return Mental.UNAGGRESSIVE;
	case "Cheerful":
		return Mental.CHEERFUL;
	case "Uncheerful":
		return Mental.UNCHEERFUL;
	case "Brave":
		return Mental.BRAVE;
	case "Unbrave":
		return Mental.UNBRAVE;
	case "Cautious":
		return Mental.CAUTIOUS;
	case "Uncautious":
		return Mental.UNCAUTIOUS;
	case "Trickish":
		return Mental.TRICKISH;
	case "Untrickish":
		return Mental.UNTRICKISH;
	default:
		throw new MotionException("Unknown mental: " ~ name);
	}
}
///精神要素を文字列へ変換。
string fromMental(Mental m) {
	switch (m) {
	case Mental.AGGRESSIVE:
		return "Aggressive";
	case Mental.UNAGGRESSIVE:
		return "Unaggressive";
	case Mental.CHEERFUL:
		return "Cheerful";
	case Mental.UNCHEERFUL:
		return "Uncheerful";
	case Mental.BRAVE:
		return "Brave";
	case Mental.UNBRAVE:
		return "Unbrave";
	case Mental.CAUTIOUS:
		return "Cautious";
	case Mental.UNCAUTIOUS:
		return "Uncautious";
	case Mental.TRICKISH:
		return "Trickish";
	case Mental.UNTRICKISH:
		return "Untrickish";
	}
}
/// 精神要素の対立側を返す。
Mental reverseMental(Mental m) {
	switch (m) {
	case Mental.AGGRESSIVE:
		return Mental.UNAGGRESSIVE;
	case Mental.UNAGGRESSIVE:
		return Mental.AGGRESSIVE;
	case Mental.CHEERFUL:
		return Mental.UNCHEERFUL;
	case Mental.UNCHEERFUL:
		return Mental.CHEERFUL;
	case Mental.BRAVE:
		return Mental.UNBRAVE;
	case Mental.UNBRAVE:
		return Mental.BRAVE;
	case Mental.CAUTIOUS:
		return Mental.UNCAUTIOUS;
	case Mental.UNCAUTIOUS:
		return Mental.CAUTIOUS;
	case Mental.TRICKISH:
		return Mental.UNTRICKISH;
	case Mental.UNTRICKISH:
		return Mental.TRICKISH;
	}
}
/// 肉体要素。
enum Physical {
	DEX, /// 敏捷度。
	AGL, /// 器用度。
	INT, /// 知力。
	STR, /// 膂力。
	VIT, /// 生命力。
	MIN, /// 精神力。
}
/// 文字列から肉体要素を生成。
Physical toPhysical(string name) {
	switch (name) {
	case "Dex":
		return Physical.DEX;
	case "Agl":
		return Physical.AGL;
	case "Int":
		return Physical.INT;
	case "Str":
		return Physical.STR;
	case "Vit":
		return Physical.VIT;
	case "Min":
		return Physical.MIN;
	default:
		throw new MotionException("Unknown physical: " ~ name);
	}
}
/// 肉体要素を文字列へ変換。
string fromPhysical(Physical p) {
	switch (p) {
	case Physical.DEX:
		return "Dex";
	case Physical.AGL:
		return "Agl";
	case Physical.INT:
		return "Int";
	case Physical.STR:
		return "Str";
	case Physical.VIT:
		return "Vit";
	case Physical.MIN:
		return "Min";
	}
}
/// 状態。
enum Status {
	ACTIVE, /// 行動可能。
	INACTIVE, /// 行動不可。
	ALIVE, /// 生存。
	DEAD, /// 非生存。
	FINE, /// 健康。
	INJURED, /// 負傷。
	HEAVY_INJURED, /// 重症。
	UNCONSCIOUS, /// 意識不明。
	POISON, /// 中毒。
	SLEEP, /// 睡眠。
	BIND, /// 呪縛。
	PARALYZE, /// 麻痺/石化。
}
/// 文字列から状態を生成。
Status toStatus(string name) {
	switch (name) {
	case "Active":
		return Status.ACTIVE;
	case "Inactive":
		return Status.INACTIVE;
	case "Alive":
		return Status.ALIVE;
	case "Dead":
		return Status.DEAD;
	case "Fine":
		return Status.FINE;
	case "Injured":
		return Status.INJURED;
	case "HeavyInjured":
		return Status.HEAVY_INJURED;
	case "Unconscious":
		return Status.UNCONSCIOUS;
	case "Poison":
		return Status.POISON;
	case "Sleep":
		return Status.SLEEP;
	case "Bind":
		return Status.BIND;
	case "Paralyze":
		return Status.PARALYZE;
	default:
		throw new MotionException("Unknown status: " ~ name);
	}
}
/// 状態を文字列へ変換。
string fromStatus(Status stat) {
	switch (stat) {
	case Status.ACTIVE:
		return "Active";
	case Status.INACTIVE:
		return "Inactive";
	case Status.ALIVE:
		return "Alive";
	case Status.DEAD:
		return "Dead";
	case Status.FINE:
		return "Fine";
	case Status.INJURED:
		return "Injured";
	case Status.HEAVY_INJURED:
		return "HeavyInjured";
	case Status.UNCONSCIOUS:
		return "Unconscious";
	case Status.POISON:
		return "Poison";
	case Status.SLEEP:
		return "Sleep";
	case Status.BIND:
		return "Bind";
	case Status.PARALYZE:
		return "Paralyze";
	}
}
/// 適用範囲。
enum Range {
	SELECTED, /// 選択中メンバ。
	RANDOM, /// 誰か一人。
	PARTY, /// パーティ全員。
	BACKPACK, /// 荷物袋。
	PARTY_AND_BACKPACK, /// 全員と荷物袋。
	FIELD, /// フィールド全体。
}
/// 文字列から適用範囲を生成。
Range toRange(string name) {
	switch (name) {
	case "Selected":
		return Range.SELECTED;
	case "Random":
		return Range.RANDOM;
	case "Party":
		return Range.PARTY;
	case "Backpack":
		return Range.BACKPACK;
	case "PartyAndBackpack":
		return Range.PARTY_AND_BACKPACK;
	case "Field":
		return Range.FIELD;
	default:
		throw new MotionException("Unknown targets: " ~ name);
	}
}
/// 適用範囲を文字列へ変換。
string fromRange(Range r) {
	switch (r) {
	case Range.SELECTED:
		return "Selected";
	case Range.RANDOM:
		return "Random";
	case Range.PARTY:
		return "Party";
	case Range.BACKPACK:
		return "Backpack";
	case Range.PARTY_AND_BACKPACK:
		return "PartyAndBackpack";
	case Range.FIELD:
		return "Field";
	}
}
/// 効果対象や話者選択時に現れる適用範囲。
Range[] RANGE_MEMBER = [Range.SELECTED, Range.RANDOM, Range.PARTY];

/// 能力修正。
enum Enhance {
	ACTION, /// 行動。
	AVOID, /// 回避。
	RESIST, /// 抵抗。
	DEFENSE, /// 防御。
}
/// 文字列から能力修正種別を生成。
Enhance toEnhance(string name) {
	switch (name) {
	case "Action":
		return Enhance.ACTION;
	case "Avoid":
		return Enhance.AVOID;
	case "Resist":
		return Enhance.RESIST;
	case "Defense":
		return Enhance.DEFENSE;
	}
}
/// 能力修正種別を文字列へ変換。
string fromEnhance(Enhance r) {
	switch (r) {
	case Enhance.ACTION:
		return "Action";
	case Enhance.AVOID:
		return "Avoid";
	case Enhance.RESIST:
		return "Resist";
	case Enhance.DEFENSE:
		return "Defense";
	}
}
/// カードの希少度。
enum Premium {
	NORMAL, /// 日用品。
	RARE, /// 希少品。
	PREMIUM, /// 貴重品。
}
/// 文字列から希少度を生成。
Premium toPremium(string name) {
	switch (name) {
	case "Normal":
		return Premium.NORMAL;
	case "Rare":
		return Premium.RARE;
	case "Premium":
		return Premium.PREMIUM;
	}
}
/// 希少度を文字列へ変換。
string fromPremium(Premium r) {
	switch (r) {
	case Premium.NORMAL:
		return "Normal";
	case Premium.RARE:
		return "Rare";
	case Premium.PREMIUM:
		return "Premium";
	}
}
/// カード効果の標的。
enum CardTarget {
	NONE, /// 対象無し。
	USER, /// 使用者。
	PARTY, /// 味方。
	ENEMY, /// 敵方。
	BOTH, /// 双方。
}
/// 文字列からカード効果標的を生成。
CardTarget toCardTarget(string name) {
	switch (name) {
	case "None":
		return CardTarget.NONE;
	case "User":
		return CardTarget.USER;
	case "Party":
		return CardTarget.PARTY;
	case "Enemy":
		return CardTarget.ENEMY;
	case "Both":
		return CardTarget.BOTH;
	}
}
/// カード効果標的を文字列へ変換。
string fromCardTarget(CardTarget r) {
	switch (r) {
	case CardTarget.NONE:
		return "None";
	case CardTarget.USER:
		return "User";
	case CardTarget.PARTY:
		return "Party";
	case CardTarget.ENEMY:
		return "Enemy";
	case CardTarget.BOTH:
		return "Both";
	}
}

/// メッセージの話者。
enum Talker {
	NARRATION, /// 話者無し。
	SELECTED, /// 選択中メンバ。
	UNSELECTED, /// 非選択メンバ。
	RANDOM, /// ランダムメンバ。
	CARD, /// カード。
	IMAGE, /// 画像ファイルを指定。
}

/// 背景遷移エフェクト。
enum Transition {
	DEFAULT, /// ユーザ指定。
	NONE, /// アニメーション無し。
	FADE, /// フェード式。
	PIXEL_DISSOLVE, /// ピクセルディゾルブ式。
	BLINDS, /// ブラインド式。
}
/// ditto
Transition[] ALL_TRANSITION = [
	Transition.DEFAULT,
	Transition.NONE,
	Transition.FADE,
	Transition.PIXEL_DISSOLVE,
	Transition.BLINDS
];
/// 文字列から背景遷移エフェクトを生成。
Transition toTransition(string name) {
	switch (name) {
	case "Default":
		return Transition.DEFAULT;
	case "None":
		return Transition.NONE;
	case "Fade":
		return Transition.FADE;
	case "PixelDissolve":
		return Transition.PIXEL_DISSOLVE;
	case "Blinds":
		return Transition.BLINDS;
	}
}
/// 背景遷移エフェクトを文字列へ変換。
string fromTransition(Transition t) {
	switch (t) {
	case Transition.DEFAULT:
		return "Default";
	case Transition.NONE:
		return "None";
	case Transition.FADE:
		return "Fade";
	case Transition.PIXEL_DISSOLVE:
		return "PixelDissolve";
	case Transition.BLINDS:
		return "Blinds";
	}
}
